﻿using System;
using System.Collections.Generic;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Reflection; // BindingFlags, PropertyInfo

namespace emanual.Wpf.Controls
{
	public partial class ColorListBox : UserControl
	{
		public static readonly DependencyProperty SelectedColorProperty;
		public static readonly RoutedEvent SelectedColorChangedEvent;

		public Color SelectedColor
		{
			get { return (Color)this.GetValue(SelectedColorProperty); }
			set { this.SetValue(SelectedColorProperty, value); }
		}

		// SelectedColor プロパティが変化したときに発生するイベントを設定・取得する
		public event RoutedPropertyChangedEventHandler<Color> SelectedColorChanged
		{
			add { this.AddHandler(SelectedColorChangedEvent, value); }
			remove { this.RemoveHandler(SelectedColorChangedEvent, value); }
		}

		//---------------------------------------------------------------------------------------------
		static ColorListBox()
		{
			SelectedColorProperty = DependencyProperty.Register("SelectedColor", typeof(Color), typeof(ColorListBox),
				new FrameworkPropertyMetadata(Colors.White, new PropertyChangedCallback(OnSelectedColorChanged)));

			SelectedColorChangedEvent = EventManager.RegisterRoutedEvent("SelectedColorChanged",
				RoutingStrategy.Bubble, typeof(RoutedPropertyChangedEventHandler<Color>), typeof(ColorListBox));
		}

		//---------------------------------------------------------------------------------------------
		public ColorListBox()
		{
			InitializeComponent();
		}

		//---------------------------------------------------------------------------------------------
		private void userControl_Loaded(object sender, RoutedEventArgs e)
		{
			this.SetupWebColorListBox();
			this.SetupNameColorListBox();
		}

		//---------------------------------------------------------------------------------------------
		private void listBox_SelectionChanged(object sender, SelectionChangedEventArgs e)
		{
			ListBox listBox = sender as ListBox;

			if (listBox != null)
			{
				int index = listBox.SelectedIndex;
				ListBoxItem item = null;

				if (listBox.Name == "lstWebColor")
				{
					item = lstWebColor.Items[index] as ListBoxItem;
				}
				else
				{
					item = lstNameColor.Items[index] as ListBoxItem;
				}

				this.SelectedColor = (Color)item.Tag;
			}
		}

		//---------------------------------------------------------------------------------------------
		private static void OnSelectedColorChanged(DependencyObject obj, DependencyPropertyChangedEventArgs args)
		{
			ColorListBox control = (ColorListBox)obj;

			RoutedPropertyChangedEventArgs<Color> e = new RoutedPropertyChangedEventArgs<Color>(
					(Color)args.OldValue, (Color)args.NewValue, SelectedColorChangedEvent);

			control.OnSelectedColorChanged(e);
		}

		//---------------------------------------------------------------------------------------------
		protected virtual void OnSelectedColorChanged(RoutedPropertyChangedEventArgs<Color> args)
		{
			this.RaiseEvent(args);
		}

		//---------------------------------------------------------------------------------------------
		// WEB カラー（色調順）リストボックスを設定する
		private void SetupWebColorListBox()
		{
			Type type = typeof(System.Windows.Media.Brushes);
			BindingFlags flags = BindingFlags.Static | BindingFlags.Public;

			PropertyInfo[] infos = type.GetProperties(flags);

			Array.Sort(infos, SortComparison);

			foreach (PropertyInfo info in infos)
			{
				SolidColorBrush brush = info.GetValue(null, null) as SolidColorBrush;

				// Transparent を除く
				if (brush == Brushes.Transparent)
					continue;

				var item = new ListBoxItem();

				var border = new Border();
				border.Background = brush;
				border.CornerRadius = new CornerRadius(2);
				border.Margin = new Thickness(10, 2, 10, 2);
				border.Width = tabControl.ActualWidth - SystemParameters.ScrollWidth - SystemParameters.BorderWidth * 2 - 30;
				border.BorderBrush = Brushes.Black;
				border.BorderThickness = new Thickness(1);

				var block = new TextBlock();
				block.Text = info.Name;
				block.Margin = new Thickness(2);
				block.HorizontalAlignment = HorizontalAlignment.Center;
				block.Foreground = this.GetTextColor(brush);

				border.Child = block;
				item.Content = border;
				item.Tag = brush.Color;

				lstWebColor.Items.Add(item);
			}
		}

		//---------------------------------------------------------------------------------------------
		// WEB カラー（名前順）リストボックスを設定する
		private void SetupNameColorListBox()
		{
			Type type = typeof(System.Windows.Media.Brushes);
			BindingFlags flags = BindingFlags.Static | BindingFlags.Public;

			PropertyInfo[] infos = type.GetProperties(flags);

			foreach (PropertyInfo info in infos)
			{
				SolidColorBrush brush = info.GetValue(null, null) as SolidColorBrush;

				// Transparent を除く
				if (brush == Brushes.Transparent)
					continue;

				var item = new ListBoxItem();

				var border = new Border();
				border.Background = brush;
				border.CornerRadius = new CornerRadius(2);
				border.Margin = new Thickness(10, 2, 10, 2);
				border.Width = tabControl.ActualWidth - SystemParameters.ScrollWidth - SystemParameters.BorderWidth * 2 - 30;
				border.BorderBrush = Brushes.Black;
				border.BorderThickness = new Thickness(1);

				var block = new TextBlock();
				block.Text = info.Name;
				block.Margin = new Thickness(2);
				block.HorizontalAlignment = HorizontalAlignment.Center;
				block.Foreground = this.GetTextColor(brush);

				border.Child = block;
				item.Content = border;
				item.Tag = brush.Color;

				lstNameColor.Items.Add(item);
			}
		}

		//-------------------------------------------------------------------------------------
		// Sort メソッドのための Comparison<T> デリゲート
		private int SortComparison(PropertyInfo info1, PropertyInfo info2)
		{
			SolidColorBrush brush1 = info1.GetValue(null, null) as SolidColorBrush;
			SolidColorBrush brush2 = info2.GetValue(null, null) as SolidColorBrush;
			Color c1 = brush1.Color;
			Color c2 = brush2.Color;

			double H1 = 0, S1 = 0, L1 = 0, H2 = 0, S2 = 0, L2 = 0;

			this.RgbToHsl(c1, ref H1, ref S1, ref L1);
			this.RgbToHsl(c2, ref H2, ref S2, ref L2);

			if (c1.A < c2.A)
			  return -1;

			if (c1.A > c2.A)
			  return 1;

			if (H1 < H2)
				return -1;

			if (H1 > H2)
				return 1;

			if (S1 < S2)
				return -1;

			if (S1 > S2)
				return 1;

			if (L1 < L2)
				return -1;

			if (L1> L2)
				return 1;

			return 0;
		}

		//---------------------------------------------------------------------------------------------
		// Color 型から HSL 値に変換する
		private void RgbToHsl(Color color, ref double hue, ref double saturation, ref double lightness)
		{
			double r, g, b, h, s, l;

			r = color.R / 255.0;
			g = color.G / 255.0;
			b = color.B / 255.0;

			double maxColor = Math.Max(r, Math.Max(g, b));
			double minColor = Math.Min(r, Math.Min(g, b));

			if (r == g && r == b)
			{
				h = 0.0;
				s = 0.0;
				l = r;
			}
			else
			{
				l = (minColor + maxColor) / 2;
				if (l < 0.5)
					s = (maxColor - minColor) / (maxColor + minColor);
				else
					s = (maxColor - minColor) / (2.0 - maxColor - minColor);

				if (r == maxColor)
					h = (g - b) / (maxColor - minColor);
				else if (g == maxColor)
					h = 2.0 + (b - r) / (maxColor - minColor);
				else
					h = 4.0 + (r - g) / (maxColor - minColor);

				h /= 6;

				if (h < 0)
					++h;
			}

			// 0 ～ 1 の範囲内に制限する
			if (h < 0) h = 0;
			if (h > 1) h = 1;
			if (s < 0) s = 0;
			if (s > 1) s = 1;
			if (l < 0) l = 0;
			if (l > 1) l = 1;

			hue = h * 360;
			saturation = s;
			lightness = l;
		}

		//---------------------------------------------------------------------------------------------
		// 背景色の明るさに応じて文字色（黒または白）を返す
		private Brush GetTextColor(SolidColorBrush brush)
		{
			Color color = brush.Color;
			Brush textBrush = Brushes.White;
			double grayScale = 0.299 * color.R + 0.587 * color.G + 0.114 * color.B;

			// 128 より明るいとき
			if (grayScale > 128)
				textBrush = Brushes.Black;

			return textBrush;
		}

	} // end of ColorListBox class
} // end of emanual.Wpf.Controls namespace
