﻿/*  AvalonEdit とともに使うユーティリティである。
 * 
 *  著作権は pmansato にあるが、コードの流用・改変は自由である。
 */
using System;
using System.Collections.Generic;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Input; // Key
using System.IO;
using ICSharpCode.AvalonEdit;
using ICSharpCode.AvalonEdit.Document;
using ICSharpCode.AvalonEdit.Editing; // TextArea
using ICSharpCode.AvalonEdit.Highlighting;
using ICSharpCode.AvalonEdit.Rendering;
using ICSharpCode.AvalonEdit.Highlighting.Xshd;
using ICSharpCode.AvalonEdit.Search; // Search
using ICSharpCode.AvalonEdit.Folding; // FoldingManager
using System.Diagnostics;

namespace emanual.Wpf.TextEditor
{
	public class SearchManager
	{
		private ICSharpCode.AvalonEdit.TextEditor FTextEditor;
		private string FSearchString = String.Empty;
		private string FReplaceString = String.Empty;
		private SearchResult FCurrentResult;
		private SearchResultBackgroundRenderer FSearchRenderer;

		public string SearchString { get { return FSearchString; } set { FSearchString = value; } }
		public string ReplaceString { get { return FReplaceString; } set { FReplaceString = value; } }

		//---------------------------------------------------------------------------------------------
		// コンストラクタ
		public SearchManager(ICSharpCode.AvalonEdit.TextEditor textEditor, SearchResultBackgroundRenderer searchRenderer)
		{
			FTextEditor = textEditor;
			FSearchRenderer = searchRenderer;
		}

		//---------------------------------------------------------------------------------------------
		// 現在のオフセット位置よりうしろの文字列を検索し、検索に成功した位置を強調表示する
		public void SearchNext()
		{
			SearchResult result;

			if (FSearchRenderer != null)
				result = FSearchRenderer.CurrentResults.FindFirstSegmentWithStartAfter(FTextEditor.TextArea.Caret.Offset + 1);
			else
				return;

			if (result == null)
				result = FSearchRenderer.CurrentResults.FirstSegment;

			if (result != null)
			{
				FCurrentResult = result;
				this.UnfoldBySearchResult(result);
			}
		}

		//---------------------------------------------------------------------------------------------
		// 現在のオフセット位置より前の文字列を検索し、検索に成功した位置を強調表示する
		public void SearchPrevious()
		{
			SearchResult result;

			if (FSearchRenderer != null)
				result = FSearchRenderer.CurrentResults.FindFirstSegmentWithStartAfter(FTextEditor.TextArea.Caret.Offset);
			else
				return;

			if (result != null)
				result = FSearchRenderer.CurrentResults.GetPreviousSegment(result);
			else
				result = FSearchRenderer.CurrentResults.LastSegment;

			if (result != null)
			{
				FCurrentResult = result;
				this.UnfoldBySearchResult(result);
			}
		}

		//---------------------------------------------------------------------------------------------
		// 現在フォーカスのある検索位置を置換する
		public void Replace()
		{
			FTextEditor.Document.Replace(FCurrentResult, FReplaceString);
		}

		//---------------------------------------------------------------------------------------------
		// 検索に成功した位置をすべて置換する
		public void ReplaceAll()
		{
			if (FSearchRenderer == null)
				return;

			if (FSearchRenderer.CurrentResults.Count < 1)
				return;

			var list = new List<SearchResult>();

			foreach (var result in FSearchRenderer.CurrentResults)
			{
				list.Add(result);	
			}

			// FSearchRenderer.CurrentResults が保持するデータは、ドキュメントの先頭からのオフセット
			// になっているので、ドキュメントの先頭から置換するとオフセットがずれてくる。そこで、
			// ドキュメントのうしろから置換する
			for (int i = list.Count - 1; i >= 0; --i)
			{
				FTextEditor.Document.Replace(list[i], FReplaceString);
			}
		}

		//---------------------------------------------------------------------------------------------
		// SearchManager を終了する（この状態は SearchManager が無効を意味する）
		public void Close(ICSharpCode.AvalonEdit.TextEditor textEditor, SearchResultBackgroundRenderer renderer)
		{
			if (renderer != null)
			{
				renderer.CurrentResults.Clear();
				FCurrentResult = null;
				textEditor.TextArea.TextView.BackgroundRenderers.Remove(renderer);
			}
		}

		//---------------------------------------------------------------------------------------------
		// FindAllAndHighlight メソッドの呼び出しによるハイライト表示をキャンセルする
		public void CancelSearchAll(ICSharpCode.AvalonEdit.TextEditor textEditor, SearchResultBackgroundRenderer renderer)
		{
			if (renderer != null)
			{
				renderer.CurrentResults.Clear();
				FCurrentResult = null;
			}
		}

		//---------------------------------------------------------------------------------------------
		// 検索に成功したすべての箇所をハイライト表示する
		public SearchResultBackgroundRenderer SearchAll(ICSharpCode.AvalonEdit.TextEditor textEditor, string searchPattern)
		{
			return SearchAll(textEditor, searchPattern, true, false, SearchMode.Normal);
		}

		//---------------------------------------------------------------------------------------------
		// 検索に成功したすべての箇所をハイライト表示する
		public SearchResultBackgroundRenderer SearchAll(ICSharpCode.AvalonEdit.TextEditor textEditor,
				string searchPattern, bool ignoreCase, bool matchWholeWords, SearchMode mode)
		{
			if (textEditor == null)
				throw new ArgumentNullException("FindAllAndHighlight()");

			if (String.IsNullOrEmpty(searchPattern))
				return null;

			TextArea textArea = textEditor.TextArea;

			FSearchRenderer = new SearchResultBackgroundRenderer();
			FSearchRenderer.MarkerBrush = textEditor.Options.SearchMarkerBrush;

			textArea.TextView.BackgroundRenderers.Add(FSearchRenderer);

			var strategy = SearchStrategyFactory.Create(searchPattern, ignoreCase, matchWholeWords, mode);

			int offset = textArea.Caret.Offset;

			foreach (SearchResult result in strategy.FindAll(textArea.Document, 0, textArea.Document.TextLength))
			{
				if (FCurrentResult == null && result.StartOffset >= offset)
				{
					FCurrentResult = result;
					this.UnfoldBySearchResult(result);  // 折りたたみ状態のテキストを展開する
				}

				FSearchRenderer.CurrentResults.Add(result);
			}

			textArea.TextView.InvalidateLayer(KnownLayer.Selection);

			return FSearchRenderer;
		}

		//---------------------------------------------------------------------------------------------
		// 検索位置が折りたたみ状態のとき、折りたたみを解除する 
		private void UnfoldBySearchResult(SearchResult result)
		{
			if ((result == null) || (FCurrentResult == null))
				return;

			TextArea textArea = FTextEditor.TextArea;

			textArea.Caret.Offset = FCurrentResult.StartOffset;
			textArea.Selection = Selection.Create(textArea, FCurrentResult.StartOffset, FCurrentResult.EndOffset);
			var foldingManager = textArea.GetService(typeof(FoldingManager)) as FoldingManager;

			if (foldingManager != null)
			{
				foreach (var folding in foldingManager.GetFoldingsContaining(result.StartOffset))
				{
					folding.IsFolded = false;
				}
			}

			textArea.Caret.BringCaretToView();
			// TextEditor にキーボードフォーカスがなくてもキャレットを表示する
			textArea.Caret.Show();
		}

		//---------------------------------------------------------------------------------------------
	} // end of SearchManager class
} // end of namespace
