﻿/*  ツールボタン用ボタンコントロールおよびトグルボタンコントロール
		
		XAML を使わず、C# コードだけでコントロールを定義するもので、このコントロールを使う場合は
		このファイルをプロジェクトに追加するだけで使用可能である。

		ボタンおよび画像のサイズはデフォルトでそれぞれ 22 x 22 と 16 x 16 になっているので、
		次のようなコードで十分な設定になる。

    xmlns:ctrl="clr-namespace:emanual.Wpf.Controls"
 
    <ctrl:ToolButtonEx x:Name="btnSave" ImageSource="images/Save.png" Click="btnSave_Click" />
 */
using System;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Media.Imaging;
//using System.Diagnostics;

namespace emanual.Wpf.Controls
{
	public class ToolButtonEx : Button
	{
		private SolidColorBrush FNormalBorderBrush = Brushes.Transparent;
		private SolidColorBrush FNormalBackgroundBrush = Brushes.Transparent;
		private SolidColorBrush FMouseOverBorderBrush = Brushes.DodgerBlue;
		private SolidColorBrush FMouseOverBackgroundBrush = Brushes.PaleTurquoise;
		private SolidColorBrush FPressedBorderBrush = Brushes.DodgerBlue;
		private SolidColorBrush FPressedBackgroundBrush = Brushes.SkyBlue;

		private DisabledImage FImage = null;

		public static readonly DependencyProperty ImageSourceProperty;
		public static readonly DependencyProperty ImageWidthProperty;
		public static readonly DependencyProperty ImageHeightProperty;

		#region property
		public ImageSource ImageSource
		{
			get { return (ImageSource)this.GetValue(ImageSourceProperty); }
			set { this.SetValue(ImageSourceProperty, value); }
		}

		public double ImageWidth
		{
			get { return (double)this.GetValue(ImageWidthProperty); }
			set { this.SetValue(ImageWidthProperty, value); }
		}

		public double ImageHeight
		{
			get { return (double)this.GetValue(ImageHeightProperty); }
			set { this.SetValue(ImageHeightProperty, value); }
		}
		#endregion
		//---------------------------------------------------------------------------------------------
		static ToolButtonEx()
		{
			ImageSourceProperty = DependencyProperty.Register("ImageSource", typeof(ImageSource), typeof(ToolButtonEx), new FrameworkPropertyMetadata(null));
			ImageWidthProperty = DependencyProperty.Register("ImageWidth", typeof(double), typeof(ToolButtonEx), new FrameworkPropertyMetadata(16.0));
			ImageHeightProperty = DependencyProperty.Register("ImageHeight", typeof(double), typeof(ToolButtonEx), new FrameworkPropertyMetadata(16.0));
		}

		//---------------------------------------------------------------------------------------------
		public ToolButtonEx()
		{
			var border = new FrameworkElementFactory(typeof(Border));
			border.Name = "PART_Border";
			border.SetValue(BorderThicknessProperty, new Thickness(1));
			border.SetValue(BackgroundProperty, FNormalBackgroundBrush);
			border.SetValue(BorderBrushProperty, FNormalBorderBrush);

			var presenter = new FrameworkElementFactory(typeof(ContentPresenter));

			var grid = new FrameworkElementFactory(typeof(Grid));

			grid.AppendChild(border);
			grid.AppendChild(presenter);

			var template = new ControlTemplate(typeof(ToolButtonEx));
			template.VisualTree = grid;

			//-----------------------------------------------------------------
			// Trigger の設定
			var mouseOverTrigger = new Trigger();
			mouseOverTrigger.Property = IsMouseOverProperty;
			mouseOverTrigger.Value = true;
			mouseOverTrigger.Setters.Add(new Setter(BorderBrushProperty, FMouseOverBorderBrush, "PART_Border"));
			mouseOverTrigger.Setters.Add(new Setter(BackgroundProperty, FMouseOverBackgroundBrush, "PART_Border"));

			var pressedTrigger = new Trigger();
			pressedTrigger.Property = IsPressedProperty;
			pressedTrigger.Value = true;
			pressedTrigger.Setters.Add(new Setter(BorderBrushProperty, FPressedBorderBrush, "PART_Border"));
			pressedTrigger.Setters.Add(new Setter(BackgroundProperty, FPressedBackgroundBrush, "PART_Border"));

			template.Triggers.Add(mouseOverTrigger);
			template.Triggers.Add(pressedTrigger);
			//-----------------------------------------------------------------

			var style = new Style(typeof(ToolButtonEx));
			style.Setters.Add(new Setter(WidthProperty, 22.0));
			style.Setters.Add(new Setter(HeightProperty, 22.0));
			style.Setters.Add(new Setter(SnapsToDevicePixelsProperty, true));
			style.Setters.Add(new Setter(OverridesDefaultStyleProperty, true));
			style.Setters.Add(new Setter(TemplateProperty, template));
			this.Style = style;

			// IsEnabled = false のときでも ToolTip を表示する
			ToolTipService.SetShowOnDisabled(this, true);
		}

		//---------------------------------------------------------------------------------------------
		protected override void OnPropertyChanged(DependencyPropertyChangedEventArgs e)
		{
			base.OnPropertyChanged(e);

			if ((e.Property == ImageSourceProperty) || (e.Property == ImageWidthProperty) || (e.Property == ImageHeightProperty))
			{
				this.UpdateImage();
			}
		}

		//---------------------------------------------------------------------------------------------
		// プロパティの変更に伴って DisabledImage コントロールを更新する
		private void UpdateImage()
		{
			FImage = new DisabledImage();
			FImage.Width = this.ImageWidth;
			FImage.Height = this.ImageHeight;
			FImage.Stretch = Stretch.Uniform;
			FImage.Source = this.ImageSource;
			FImage.VerticalAlignment = VerticalAlignment.Center;
			FImage.HorizontalAlignment = HorizontalAlignment.Center;

			this.Content = FImage;
		}
		//---------------------------------------------------------------------------------------------
	} // end of ToolButtonEx class

	//***********************************************************************************************
	// ToolButtonEx コントロールの IsChecked プロパティに対応するトグルボタンコントロール
	public class ToolToggleButtonEx : System.Windows.Controls.Primitives.ToggleButton
	{
		private SolidColorBrush FNormalBorderBrush = Brushes.Transparent;
		private SolidColorBrush FNormalBackgroundBrush = Brushes.Transparent;
		private SolidColorBrush FMouseOverBorderBrush = Brushes.DodgerBlue;
		private SolidColorBrush FMouseOverBackgroundBrush = Brushes.PaleTurquoise;
		private SolidColorBrush FPressedBorderBrush = Brushes.DodgerBlue;
		private SolidColorBrush FPressedBackgroundBrush = Brushes.SkyBlue;
		private SolidColorBrush FCheckedBorderBrush = Brushes.Orange;
		private SolidColorBrush FCheckedBackgroundBrush = Brushes.Wheat;

		private DisabledImage FImage = null;

		public static readonly DependencyProperty ImageSourceProperty;
		public static readonly DependencyProperty ImageWidthProperty;
		public static readonly DependencyProperty ImageHeightProperty;

		#region property
		public ImageSource ImageSource
		{
			get { return (ImageSource)this.GetValue(ImageSourceProperty); }
			set { this.SetValue(ImageSourceProperty, value); }
		}

		public double ImageWidth
		{
			get { return (double)this.GetValue(ImageWidthProperty); }
			set { this.SetValue(ImageWidthProperty, value); }
		}

		public double ImageHeight
		{
			get { return (double)this.GetValue(ImageHeightProperty); }
			set { this.SetValue(ImageHeightProperty, value); }
		}
		#endregion
		//---------------------------------------------------------------------------------------------
		static ToolToggleButtonEx()
		{
			ImageSourceProperty = DependencyProperty.Register("ImageSource", typeof(ImageSource), typeof(ToolToggleButtonEx), new FrameworkPropertyMetadata(null));
			ImageWidthProperty = DependencyProperty.Register("ImageWidth", typeof(double), typeof(ToolToggleButtonEx), new FrameworkPropertyMetadata(16.0));
			ImageHeightProperty = DependencyProperty.Register("ImageHeight", typeof(double), typeof(ToolToggleButtonEx), new FrameworkPropertyMetadata(16.0));
		}

		//---------------------------------------------------------------------------------------------
		public ToolToggleButtonEx()
		{
			var border = new FrameworkElementFactory(typeof(Border));
			border.Name = "PART_Border";
			border.SetValue(BorderThicknessProperty, new Thickness(1));
			border.SetValue(BackgroundProperty, FNormalBackgroundBrush);
			border.SetValue(BorderBrushProperty, FNormalBorderBrush);

			var presenter = new FrameworkElementFactory(typeof(ContentPresenter));

			var grid = new FrameworkElementFactory(typeof(Grid));

			grid.AppendChild(border);
			grid.AppendChild(presenter);

			var template = new ControlTemplate(typeof(ToolToggleButtonEx));
			template.VisualTree = grid;

			//-----------------------------------------------------------------
			// Trigger の設定
			var mouseOverTrigger = new Trigger();
			mouseOverTrigger.Property = IsMouseOverProperty;
			mouseOverTrigger.Value = true;
			mouseOverTrigger.Setters.Add(new Setter(BorderBrushProperty, FMouseOverBorderBrush, "PART_Border"));
			mouseOverTrigger.Setters.Add(new Setter(BackgroundProperty, FMouseOverBackgroundBrush, "PART_Border"));

			var pressedTrigger = new Trigger();
			pressedTrigger.Property = IsPressedProperty;
			pressedTrigger.Value = true;
			pressedTrigger.Setters.Add(new Setter(BorderBrushProperty, FPressedBorderBrush, "PART_Border"));
			pressedTrigger.Setters.Add(new Setter(BackgroundProperty, FPressedBackgroundBrush, "PART_Border"));

			var checkedTrigger = new Trigger();
			checkedTrigger.Property = IsCheckedProperty;
			checkedTrigger.Value = true;
			checkedTrigger.Setters.Add(new Setter(BorderBrushProperty, FCheckedBorderBrush, "PART_Border"));
			checkedTrigger.Setters.Add(new Setter(BackgroundProperty, FCheckedBackgroundBrush, "PART_Border"));

			template.Triggers.Add(mouseOverTrigger);
			template.Triggers.Add(pressedTrigger);
			template.Triggers.Add(checkedTrigger);
			//-----------------------------------------------------------------

			var style = new Style(typeof(ToolToggleButtonEx));
			style.Setters.Add(new Setter(WidthProperty, 22.0));
			style.Setters.Add(new Setter(HeightProperty, 22.0));
			style.Setters.Add(new Setter(SnapsToDevicePixelsProperty, true));
			style.Setters.Add(new Setter(OverridesDefaultStyleProperty, true));
			style.Setters.Add(new Setter(TemplateProperty, template));
			this.Style = style;

			// IsEnabled = false のときでも ToolTip を表示する
			ToolTipService.SetShowOnDisabled(this, true);
		}

		//---------------------------------------------------------------------------------------------
		protected override void OnPropertyChanged(DependencyPropertyChangedEventArgs e)
		{
			base.OnPropertyChanged(e);

			if ((e.Property == ImageSourceProperty) || (e.Property == ImageWidthProperty) || (e.Property == ImageHeightProperty))
			{
				this.UpdateImage();
			}
		}

		//---------------------------------------------------------------------------------------------
		// プロパティの変更に伴って DisabledImage コントロールを更新する
		private void UpdateImage()
		{
			FImage = new DisabledImage();
			FImage.Width = this.ImageWidth;
			FImage.Height = this.ImageHeight;
			FImage.Stretch = Stretch.Uniform;
			FImage.Source = this.ImageSource;
			FImage.VerticalAlignment = VerticalAlignment.Center;
			FImage.HorizontalAlignment = HorizontalAlignment.Center;

			this.Content = FImage;
		}
		//---------------------------------------------------------------------------------------------
	} // end of ToolToggleButtonEx class

	//***********************************************************************************************
	// IsEnabled プロパティが false のとき、画像をグレースケール表示する Image コントロール
	// 単独でも使用可能だが、通常はツールバー用ボタンとかメニュー項目で使用する
	public class DisabledImage : Image
	{
		static DisabledImage()
		{
			IsEnabledProperty.OverrideMetadata(typeof(DisabledImage), new FrameworkPropertyMetadata(true, new PropertyChangedCallback(OnIsEnabledPropertyChanged)));
		}

		private static void OnIsEnabledPropertyChanged(DependencyObject obj, DependencyPropertyChangedEventArgs args)
		{
			var image = obj as DisabledImage;

			if ((image == null) || (image.Source == null))
				return;

			var isEnabled = Convert.ToBoolean(args.NewValue);

			if (isEnabled == false)
			{
				var bitmap = image.Source.Clone();
				image.Source = new FormatConvertedBitmap((BitmapSource)bitmap, PixelFormats.Gray32Float, null, 0);
				// FormatConvertedBitmap はアルファ値を維持しないので、背景に対して透過的に描画するためにマスクを設定する
				image.OpacityMask = new ImageBrush(bitmap);
			}
			else
			{
				image.Source = ((FormatConvertedBitmap)image.Source).Source;
				image.OpacityMask = null;
			}
		}
	} // end of DisabledImage class
} // end of emanual.Wpf.Controls namespace
