﻿// Copyright (c) AlphaSierraPapa for the SharpDevelop Team (for details please see \doc\copyright.txt)
// This code is distributed under the GNU LGPL (for details please see \doc\license.txt)

using System;
using System.Diagnostics;
using System.Globalization;

namespace ICSharpCode.AvalonEdit.Document
{
	// Represents a line inside a <see cref="TextDocument"/>.
	// The <see cref="TextDocument.Lines"/> collection contains one DocumentLine instance
	// for every line in the document. This collection is read-only to user code and is automatically
	// updated to reflect the current document content.
	// Internally, the DocumentLine instances are arranged in a binary tree that allows for both efficient updates and lookup.
	// Converting between offset and line number is possible in O(lg N) time,
	// and the data structure also updates all offsets in O(lg N) whenever a line is inserted or removed.
	public sealed partial class DocumentLine : ISegment
	{
#if DEBUG
		// Required for thread safety check which is done only in debug builds.
		// To save space, we don't store the document reference in release builds as we don't need it there.
		readonly TextDocument document;
#endif

		internal bool isDeleted;

		internal DocumentLine(TextDocument document)
		{
#if DEBUG
			Debug.Assert(document != null);
			this.document = document;
#endif
		}

		[Conditional("DEBUG")]
		void DebugVerifyAccess()
		{
#if DEBUG
			document.DebugVerifyAccess();
#endif
		}

		// Gets if this line was deleted from the document.
		public bool IsDeleted
		{
			get
			{
				DebugVerifyAccess();
				return isDeleted;
			}
		}

		// Gets the number of this line.
		// Runtime: O(log n)
		public int LineNumber
		{
			get
			{
				if (IsDeleted)
					throw new InvalidOperationException();
				return DocumentLineTree.GetIndexFromNode(this) + 1;
			}
		}

		// Gets the starting offset of the line in the document's text.
		// Runtime: O(log n)
		public int Offset
		{
			get
			{
				if (IsDeleted)
					throw new InvalidOperationException();
				return DocumentLineTree.GetOffsetFromNode(this);
			}
		}

		// Gets the end offset of the line in the document's text (the offset before the line delimiter).
		// Runtime: O(log n)
		// <exception cref="InvalidOperationException">The line was deleted.</exception>
		// <remarks>EndOffset = <see cref="Offset"/> + <see cref="Length"/>.</remarks>
		public int EndOffset
		{
			get { return this.Offset + this.Length; }
		}

		int totalLength;
		byte delimiterLength;

		// Gets the length of this line. The length does not include the line delimiter. O(1)
		// <remarks>This property is still available even if the line was deleted;
		// in that case, it contains the line's length before the deletion.</remarks>
		public int Length
		{
			get
			{
				DebugVerifyAccess();
				return totalLength - delimiterLength;
			}
		}

		// Gets the length of this line, including the line delimiter. O(1)
		// <remarks>This property is still available even if the line was deleted;
		// in that case, it contains the line's length before the deletion.</remarks>
		public int TotalLength
		{
			get
			{
				DebugVerifyAccess();
				return totalLength;
			}
			internal set
			{
				// this is set by DocumentLineTree
				totalLength = value;
			}
		}

		// <para>Gets the length of the line delimiter.</para>
		// <para>The value is 1 for single <c>"\r"</c> or <c>"\n"</c>, 2 for the <c>"\r\n"</c> sequence;
		// and 0 for the last line in the document.</para>
		// <remarks>This property is still available even if the line was deleted;
		// in that case, it contains the line delimiter's length before the deletion.</remarks>
		public int DelimiterLength
		{
			get
			{
				DebugVerifyAccess();
				return delimiterLength;
			}
			internal set
			{
				Debug.Assert(value >= 0 && value <= 2);
				delimiterLength = (byte)value;
			}
		}

		// Gets the next line in the document.
		public DocumentLine NextLine
		{
			get
			{
				DebugVerifyAccess();

				if (right != null)
				{
					return right.LeftMost;
				}
				else
				{
					DocumentLine node = this;
					DocumentLine oldNode;
					do
					{
						oldNode = node;
						node = node.parent;
						// we are on the way up from the right part, don't output node again
					} while (node != null && node.right == oldNode);
					return node;
				}
			}
		}

		// Gets the previous line in the document.
		public DocumentLine PreviousLine
		{
			get
			{
				DebugVerifyAccess();

				if (left != null)
				{
					return left.RightMost;
				}
				else
				{
					DocumentLine node = this;
					DocumentLine oldNode;
					do
					{
						oldNode = node;
						node = node.parent;
						// we are on the way up from the left part, don't output node again
					} while (node != null && node.left == oldNode);
					return node;
				}
			}
		}

		// Gets a string with debug output showing the line number and offset.
		// Does not include the line's text.
		public override string ToString()
		{
			if (IsDeleted)
				return "[DocumentLine deleted]";
			else
				return string.Format(
					CultureInfo.InvariantCulture,
					"[DocumentLine Number={0} Offset={1} Length={2}]", LineNumber, Offset, Length);
		}
	}
}
