﻿// Copyright (c) AlphaSierraPapa for the SharpDevelop Team (for details please see \doc\copyright.txt)
// This code is distributed under the GNU LGPL (for details please see \doc\license.txt)

using System;
using System.Diagnostics;
using System.Windows;

using ICSharpCode.AvalonEdit.Document;
using ICSharpCode.AvalonEdit.Rendering;

namespace ICSharpCode.AvalonEdit.Editing
{
	// Base class for margins.
	// Margins don't have to derive from this class, it just helps maintaining a reference to the TextView and the TextDocument.
	// AbstractMargin derives from FrameworkElement, so if you don't want to handle visual children and rendering
	// on your own, choose another base class for your margin!
	public abstract class AbstractMargin : FrameworkElement, ITextViewConnect
	{
		// automatically set/unset TextView property using ITextViewConnect
		bool wasAutoAddedToTextView;
		TextDocument document;

		public static readonly DependencyProperty TextViewProperty = DependencyProperty.Register(
			"TextView", typeof(TextView), typeof(AbstractMargin), new FrameworkPropertyMetadata(OnTextViewChanged));

		// Gets the document associated with the margin.
		public TextDocument Document
		{
			get { return document; }
		}

		// Gets/sets the text view for which line numbers are displayed.
		// remarks : Adding a margin to "TextArea.LeftMargins" will automatically set this property to the text area's TextView.
		public TextView TextView
		{
			get { return (TextView)GetValue(TextViewProperty); }
			set { SetValue(TextViewProperty, value); }
		}

		static void OnTextViewChanged(DependencyObject dp, DependencyPropertyChangedEventArgs e)
		{
			AbstractMargin margin = (AbstractMargin)dp;
			margin.wasAutoAddedToTextView = false;
			margin.OnTextViewChanged((TextView)e.OldValue, (TextView)e.NewValue);
		}

		void ITextViewConnect.AddToTextView(TextView textView)
		{
			if (this.TextView == null)
			{
				this.TextView = textView;
				wasAutoAddedToTextView = true;
			}
			else if (this.TextView != textView)
			{
				throw new InvalidOperationException("This margin belongs to a different TextView.");
			}
		}

		void ITextViewConnect.RemoveFromTextView(TextView textView)
		{
			if (wasAutoAddedToTextView && this.TextView == textView)
			{
				this.TextView = null;
				Debug.Assert(!wasAutoAddedToTextView); // setting this.TextView should have unset this flag
			}
		}

		// Called when the <see cref="TextView"/> is changing.
		protected virtual void OnTextViewChanged(TextView oldTextView, TextView newTextView)
		{
			if (oldTextView != null)
			{
				oldTextView.DocumentChanged -= TextViewDocumentChanged;
			}
			if (newTextView != null)
			{
				newTextView.DocumentChanged += TextViewDocumentChanged;
			}
			TextViewDocumentChanged(null, null);
		}

		void TextViewDocumentChanged(object sender, EventArgs e)
		{
			OnDocumentChanged(document, TextView != null ? TextView.Document : null);
		}

		// Called when the <see cref="Document"/> is changing.
		protected virtual void OnDocumentChanged(TextDocument oldDocument, TextDocument newDocument)
		{
			document = newDocument;
		}
	}
}
