﻿// Copyright (c) AlphaSierraPapa for the SharpDevelop Team (for details please see \doc\copyright.txt)
// This code is distributed under the GNU LGPL (for details please see \doc\license.txt)

using System;
using System.Collections.Generic;
using ICSharpCode.AvalonEdit.Document;

namespace ICSharpCode.AvalonEdit.Editing
{
	// Implementation for <see cref="IReadOnlySectionProvider"/> that stores the segments in a "TextSegmentCollection{T}".
	public class TextSegmentReadOnlySectionProvider<T> : IReadOnlySectionProvider where T : TextSegment
	{
		readonly TextSegmentCollection<T> segments;

		// Gets the collection storing the read-only segments.
		public TextSegmentCollection<T> Segments
		{
			get { return segments; }
		}

		// Creates a new TextSegmentReadOnlySectionProvider instance for the specified document.
		public TextSegmentReadOnlySectionProvider(TextDocument textDocument)
		{
			segments = new TextSegmentCollection<T>(textDocument);
		}

		// Creates a new TextSegmentReadOnlySectionProvider instance using the specified TextSegmentCollection.
		public TextSegmentReadOnlySectionProvider(TextSegmentCollection<T> segments)
		{
			if (segments == null)
				throw new ArgumentNullException("segments");
			this.segments = segments;
		}

		// Gets whether insertion is possible at the specified offset.
		public virtual bool CanInsert(int offset)
		{
			foreach (TextSegment segment in segments.FindSegmentsContaining(offset))
			{
				if (segment.StartOffset < offset && offset < segment.EndOffset)
					return false;
			}
			return true;
		}

		// Gets the deletable segments inside the given segment.
		public virtual IEnumerable<ISegment> GetDeletableSegments(ISegment segment)
		{
			if (segment == null)
				throw new ArgumentNullException("segment");

			int readonlyUntil = segment.Offset;

			foreach (TextSegment ts in segments.FindOverlappingSegments(segment))
			{
				int start = ts.StartOffset;
				int end = start + ts.Length;
				if (start > readonlyUntil)
				{
					yield return new SimpleSegment(readonlyUntil, start - readonlyUntil);
				}
				if (end > readonlyUntil)
				{
					readonlyUntil = end;
				}
			}

			int endOffset = segment.EndOffset;

			if (readonlyUntil < endOffset)
			{
				yield return new SimpleSegment(readonlyUntil, endOffset - readonlyUntil);
			}
		}
	}
}
