﻿// Copyright (c) AlphaSierraPapa for the SharpDevelop Team (for details please see \doc\copyright.txt)
// This code is distributed under the GNU LGPL (for details please see \doc\license.txt)

using System;
using System.Collections.Generic;

namespace ICSharpCode.AvalonEdit.Rendering
{
	// Base class for "IVisualLineTransformer" that helps
	// splitting visual elements so that colors (and other text properties) can be easily assigned
	// to individual words/characters.
	public abstract class ColorizingTransformer : IVisualLineTransformer, ITextViewConnect
	{
		// Gets the list of elements currently being transformed.
		protected IList<VisualLineElement> CurrentElements { get; private set; }

		// "IVisualLineTransformer.Transform" implementation.
		// Sets "CurrentElements" and calls "Colorize".
		public void Transform(ITextRunConstructionContext context, IList<VisualLineElement> elements)
		{
			if (elements == null)
				throw new ArgumentNullException("elements");
			if (this.CurrentElements != null)
				throw new InvalidOperationException("Recursive Transform() call");
			this.CurrentElements = elements;

			try
			{
				Colorize(context);
			}
			finally
			{
				this.CurrentElements = null;
			}
		}

		// Performs the colorization.
		protected abstract void Colorize(ITextRunConstructionContext context);

		// Changes visual element properties.
		// This method accesses "CurrentElements", so it must be called only during a "Transform" call.
		// This method splits "VisualLineElement"s as necessary to ensure that the region
		// can be colored by setting the "VisualLineElement.TextRunProperties" of whole elements,
		// and then calls the "action" on all elements in the region.
		// visualStartColumn : Start visual column of the region to change
		// visualEndColumn   : End visual column of the region to change
		// action            : Action that changes an individual "VisualLineElement".
		protected void ChangeVisualElements(int visualStartColumn, int visualEndColumn, Action<VisualLineElement> action)
		{
			if (action == null)
				throw new ArgumentNullException("action");

			for (int i = 0; i < CurrentElements.Count; i++)
			{
				VisualLineElement e = CurrentElements[i];
				if (e.VisualColumn > visualEndColumn)
					break;

				if (e.VisualColumn < visualStartColumn && e.VisualColumn + e.VisualLength > visualStartColumn)
				{
					if (e.CanSplit)
					{
						e.Split(visualStartColumn, CurrentElements, i--);
						continue;
					}
				}

				if (e.VisualColumn >= visualStartColumn && e.VisualColumn < visualEndColumn)
				{
					if (e.VisualColumn + e.VisualLength > visualEndColumn)
					{
						if (e.CanSplit)
						{
							e.Split(visualEndColumn, CurrentElements, i--);
							continue;
						}
					}
					else
					{
						action(e);
					}
				}
			}
		}

		// Called when added to a text view.
		protected virtual void OnAddToTextView(TextView textView)
		{
		}

		// Called when removed from a text view.
		protected virtual void OnRemoveFromTextView(TextView textView)
		{
		}

		void ITextViewConnect.AddToTextView(TextView textView)
		{
			OnAddToTextView(textView);
		}

		void ITextViewConnect.RemoveFromTextView(TextView textView)
		{
			OnRemoveFromTextView(textView);
		}
	}
}
