﻿// Copyright (c) AlphaSierraPapa for the SharpDevelop Team (for details please see \doc\copyright.txt)
// This code is distributed under the GNU LGPL (for details please see \doc\license.txt)

using System;
using System.ComponentModel;
using System.Reflection;
using System.Text;

namespace ICSharpCode.AvalonEdit
{
	// A container for the text editor options.
	[Serializable]
	public class TextEditorOptions : INotifyPropertyChanged
	{
		bool allowScrollBelowDocument;
		bool cutCopyWholeLine = true;
		bool enableEmailHyperlinks = true;
		bool enableHyperlinks = true;
		bool enableImeSupport = true;
		bool enableRectangularSelection = true;
		bool enableTextDragDrop = true;
		bool enableVirtualSpace;
		bool inheritWordWrapIndentation = true;
		bool requireControlModifierForHyperlinkClick = true;
		bool showBoxForControlCharacters = true;
		bool showColumnRuler = false; //<< 4.3 に追加
		bool showEndOfLine;
		bool showSpaces;
		bool showTabs;
		int columnRulerPosition = 80; //<< 4.3 に追加
		double wordWrapIndentation = 0;

		// I'm using '_' prefixes for the fields here to avoid confusion with the local variables in the methods below.
		// The fields should be accessed only by their property - the fields might not be used if someone overrides the property.
		int _indentationSize = 4;
		bool _convertTabsToSpaces;

		[field: NonSerialized]
		public event PropertyChangedEventHandler PropertyChanged;

		//<< 以下の 3 つのプロパティは独自に追加（デフォルトはオリジナルのままである）
		private System.Windows.Media.Brush searchMarkerBrush = System.Windows.Media.Brushes.LightGreen;
		private double markerCornerRadius = 3.0;
		private string newLine = "\u00B6";

		#region property
		// Gets/Sets the brush for search marker brush.
		public System.Windows.Media.Brush SearchMarkerBrush //<< 独自に追加
		{
			get { return searchMarkerBrush; }
			set
			{
				if (searchMarkerBrush != value)
				{
					searchMarkerBrush = value;
					OnPropertyChanged("SearchMarkerBrush");
				}
			}
		}

		// Get/Set search marker corner radius
		public double MarkerCornerRadius //<< 独自に追加
		{
			get { return markerCornerRadius; }
			set { markerCornerRadius = value; }
		}

		// Gets/Sets the new line string.
		public string NewLine //<< 独自に追加
		{
			get { return newLine; }
			set
			{
				if (newLine != value)
				{
					newLine = value;
					OnPropertyChanged("NewLine");
				}
			}
		}

		// Gets/Sets whether to show · for spaces.
		// remarks : The default value is false.
		[DefaultValue(false)]
		public virtual bool ShowSpaces
		{
			get { return showSpaces; }
			set
			{
				if (showSpaces != value)
				{
					showSpaces = value;
					OnPropertyChanged("ShowSpaces");
				}
			}
		}

		// Gets/Sets whether to show » for tabs.
		// remarks : The default value is false.
		[DefaultValue(false)]
		public virtual bool ShowTabs
		{
			get { return showTabs; }
			set
			{
				if (showTabs != value)
				{
					showTabs = value;
					OnPropertyChanged("ShowTabs");
				}
			}
		}

		// Gets/Sets whether to show ¶ at the end of lines.
		// remarks : The default value is false.
		[DefaultValue(false)]
		public virtual bool ShowEndOfLine
		{
			get { return showEndOfLine; }
			set
			{
				if (showEndOfLine != value)
				{
					showEndOfLine = value;
					OnPropertyChanged("ShowEndOfLine");
				}
			}
		}

		// Gets/Sets whether to show a box with the hex code for control characters.
		// remarks : The default value is true.
		[DefaultValue(true)]
		public virtual bool ShowBoxForControlCharacters
		{
			get { return showBoxForControlCharacters; }
			set
			{
				if (showBoxForControlCharacters != value)
				{
					showBoxForControlCharacters = value;
					OnPropertyChanged("ShowBoxForControlCharacters");
				}
			}
		}

		// Gets/Sets whether to enable clickable hyperlinks in the editor.
		// remarks : The default value is true.
		[DefaultValue(true)]
		public virtual bool EnableHyperlinks
		{
			get { return enableHyperlinks; }
			set
			{
				if (enableHyperlinks != value)
				{
					enableHyperlinks = value;
					OnPropertyChanged("EnableHyperlinks");
				}
			}
		}

		// Gets/Sets whether to enable clickable hyperlinks for e-mail addresses in the editor.
		// remarks : The default value is true.
		[DefaultValue(true)]
		public virtual bool EnableEmailHyperlinks
		{
			get { return enableEmailHyperlinks; }
			set
			{
				if (enableEmailHyperlinks != value)
				{
					enableEmailHyperlinks = value;
					OnPropertyChanged("EnableEMailHyperlinks");
				}
			}
		}

		// Gets/Sets whether the user needs to press Control to click hyperlinks.
		// The default value is true.
		// remarks : The default value is true.
		[DefaultValue(true)]
		public virtual bool RequireControlModifierForHyperlinkClick
		{
			get { return requireControlModifierForHyperlinkClick; }
			set
			{
				if (requireControlModifierForHyperlinkClick != value)
				{
					requireControlModifierForHyperlinkClick = value;
					OnPropertyChanged("RequireControlModifierForHyperlinkClick");
				}
			}
		}

		// Gets/Sets the width of one indentation unit.
		// remarks : The default value is 4.
		[DefaultValue(4)]
		public virtual int IndentationSize
		{
			get { return _indentationSize; }
			set
			{
				if (value < 1)
					throw new ArgumentOutOfRangeException("value", value, "value must be positive");
				// sanity check; a too large value might cause WPF to crash internally much later
				// (it only crashed in the hundred thousands for me; but might crash earlier with larger fonts)
				if (value > 1000)
					throw new ArgumentOutOfRangeException("value", value, "indentation size is too large");

				if (_indentationSize != value)
				{
					_indentationSize = value;
					OnPropertyChanged("IndentationSize");
					OnPropertyChanged("IndentationString");
				}
			}
		}

		// Gets/Sets whether to use spaces for indentation instead of tabs.
		// remarks : The default value is false.
		[DefaultValue(false)]
		public virtual bool ConvertTabsToSpaces
		{
			get { return _convertTabsToSpaces; }
			set
			{
				if (_convertTabsToSpaces != value)
				{
					_convertTabsToSpaces = value;
					OnPropertyChanged("ConvertTabsToSpaces");
					OnPropertyChanged("IndentationString");
				}
			}
		}

		// Gets the text used for indentation.
		[System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Naming", "CA1721:PropertyNamesShouldNotMatchGetMethods")]
		[Browsable(false)]
		public string IndentationString
		{
			get { return GetIndentationString(1); }
		}

		// Gets/Sets whether copying without a selection copies the whole current line.
		[DefaultValue(true)]
		public virtual bool CutCopyWholeLine
		{
			get { return cutCopyWholeLine; }
			set
			{
				if (cutCopyWholeLine != value)
				{
					cutCopyWholeLine = value;
					OnPropertyChanged("CutCopyWholeLine");
				}
			}
		}

		// Gets/Sets whether the user can scroll below the bottom of the document.
		// The default value is false; but it a good idea to set this property to true when using folding.
		[DefaultValue(false)]
		public virtual bool AllowScrollBelowDocument
		{
			get { return allowScrollBelowDocument; }
			set
			{
				if (allowScrollBelowDocument != value)
				{
					allowScrollBelowDocument = value;
					OnPropertyChanged("AllowScrollBelowDocument");
				}
			}
		}

		// Gets/Sets the indentation used for all lines except the first when word-wrapping.
		// The default value is 0.
		[DefaultValue(0.0)]
		public virtual double WordWrapIndentation
		{
			get { return wordWrapIndentation; }
			set
			{
				if (double.IsNaN(value) || double.IsInfinity(value))
					throw new ArgumentOutOfRangeException("value", value, "value must not be NaN/infinity");
				if (value != wordWrapIndentation)
				{
					wordWrapIndentation = value;
					OnPropertyChanged("WordWrapIndentation");
				}
			}
		}

		// Gets/Sets whether the indentation is inherited from the first line when word-wrapping.
		// The default value is true.
		// remarks : When combined with "WordWrapIndentation" the inherited indentation is added to the word wrap indentation.
		[DefaultValue(true)]
		public virtual bool InheritWordWrapIndentation
		{
			get { return inheritWordWrapIndentation; }
			set
			{
				if (value != inheritWordWrapIndentation)
				{
					inheritWordWrapIndentation = value;
					OnPropertyChanged("InheritWordWrapIndentation");
				}
			}
		}

		// Enables rectangular selection (press ALT and select a rectangle)
		[DefaultValue(true)]
		public bool EnableRectangularSelection
		{
			get { return enableRectangularSelection; }
			set
			{
				if (enableRectangularSelection != value)
				{
					enableRectangularSelection = value;
					OnPropertyChanged("EnableRectangularSelection");
				}
			}
		}

		// Enable dragging text within the text area.
		[DefaultValue(true)]
		public bool EnableTextDragDrop
		{
			get { return enableTextDragDrop; }
			set
			{
				if (enableTextDragDrop != value)
				{
					enableTextDragDrop = value;
					OnPropertyChanged("EnableTextDragDrop");
				}
			}
		}

		// Gets/Sets whether the user can set the caret behind the line ending (into "virtual space").
		// Note that virtual space is always used (independent from this setting)
		// when doing rectangle selections.
		[DefaultValue(false)]
		public virtual bool EnableVirtualSpace
		{
			get { return enableVirtualSpace; }
			set
			{
				if (enableVirtualSpace != value)
				{
					enableVirtualSpace = value;
					OnPropertyChanged("EnableVirtualSpace");
				}
			}
		}

		// Gets/Sets whether the support for Input Method Editors (IME)
		// for non-alphanumeric scripts (Chinese, Japanese, Korean, ...) is enabled.
		[DefaultValue(true)]
		public virtual bool EnableImeSupport
		{
			get { return enableImeSupport; }
			set
			{
				if (enableImeSupport != value)
				{
					enableImeSupport = value;
					OnPropertyChanged("EnableImeSupport");
				}
			}
		}

		// Gets/Sets whether the column ruler should be shown.
		[DefaultValue(false)]
		public virtual bool ShowColumnRuler
		{
			get { return showColumnRuler; }
			set
			{
				if (showColumnRuler != value)
				{
					showColumnRuler = value;
					OnPropertyChanged("ShowColumnRuler");
				}
			}
		}

		// Gets/Sets where the column ruler should be shown.
		[DefaultValue(80)]
		public virtual int ColumnRulerPosition
		{
			get { return columnRulerPosition; }
			set
			{
				if (columnRulerPosition != value)
				{
					columnRulerPosition = value;
					OnPropertyChanged("ColumnRulerPosition");
				}
			}
		}
		#endregion
		//---------------------------------------------------------------------------------------------
		// Initializes an empty instance of TextEditorOptions.
		public TextEditorOptions()
		{
		}

		// Initializes a new instance of TextEditorOptions by copying all values from "options"to the new instance.
		public TextEditorOptions(TextEditorOptions options)
		{
			// get all the fields in the class
			FieldInfo[] fields = typeof(TextEditorOptions).GetFields(BindingFlags.NonPublic | BindingFlags.Public | BindingFlags.Instance);

			// copy each value over to 'this'
			foreach (FieldInfo fi in fields)
			{
				if (!fi.IsNotSerialized)
					fi.SetValue(this, fi.GetValue(options));
			}
		}

		// Raises the PropertyChanged event.
		// propertyName : The name of the changed property.
		protected void OnPropertyChanged(string propertyName)
		{
			OnPropertyChanged(new PropertyChangedEventArgs(propertyName));
		}

		// Raises the PropertyChanged event.
		protected virtual void OnPropertyChanged(PropertyChangedEventArgs e)
		{
			if (PropertyChanged != null)
			{
				PropertyChanged(this, e);
			}
		}

		// Gets text required to indent from the specified <paramref name="column"/> to the next indentation level.
		public virtual string GetIndentationString(int column)
		{
			if (column < 1)
				throw new ArgumentOutOfRangeException("column", column, "Value must be at least 1.");

			int indentationSize = this.IndentationSize;

			if (ConvertTabsToSpaces)
			{
				return new string(' ', indentationSize - ((column - 1) % indentationSize));
			}
			else
			{
				return "\t";
			}
		}

	} // end of TextEditorOptions class
} // end of ICSharpCode.AvalonEdit namespace
