﻿using System;
using System.Collections.Generic;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.ComponentModel;
using System.Collections.ObjectModel;
using System.Xml;
using System.Xml.XPath;
using System.Reflection;
using ICSharpCode.AvalonEdit;
using ICSharpCode.AvalonEdit.Highlighting;
using ICSharpCode.AvalonEdit.Document;
using System.Diagnostics;

namespace emanual.Wpf.TextEditor
{
	public partial class MainWindow : Window
	{
		private const string INI_FILE = "XshdEditor.xmlini";
		private const string GENERAL_SEC = "General";
		private const string FORM_POS_KEY = "FormPosition";
		private const string BACKGROUND_KEY = "Background";
		private const string FOREGROUND_KEY = "Foreground";

		private string FInitialDIR = String.Empty;
		private ObservableCollection<XshdData> FDataCollection;
		private ColorPopup FForegroundPopup;
		private ColorPopup FBackgroundPopup;
		private int FButtonTag;

		private XmlDocument FDocument = null;
		private string FXshdFileName = String.Empty;
		private SolidColorBrush FEditorBackground = null; // コントロールの背景色
		private SolidColorBrush FEditorForeground = null; // 通常表示の文字色

		//---------------------------------------------------------------------------------------------
		public MainWindow()
		{
			InitializeComponent();
		}

		//---------------------------------------------------------------------------------------------
		private void Window_Loaded(object sender, RoutedEventArgs e)
		{
			// 前回起動時のフォームの位置とサイズを再現する
			var ini = new emanual.Wpf.Utility.XmlIni(INI_FILE);
			ini.SetFormPositionAndSize(this, GENERAL_SEC, FORM_POS_KEY, 100, 100, 700, 400);

			FEditorBackground = (SolidColorBrush)ini.ReadValue(GENERAL_SEC, BACKGROUND_KEY, SystemColors.WindowBrush);
			FEditorForeground = (SolidColorBrush)ini.ReadValue(GENERAL_SEC, FOREGROUND_KEY, SystemColors.ControlTextBrush);

			if (FEditorBackground == null)
				FEditorBackground = SystemColors.WindowBrush;

			if (FEditorForeground == null)
				FEditorForeground = SystemColors.ControlTextBrush;

			FForegroundPopup = new ColorPopup();
			FForegroundPopup.Closed += new EventHandler(foregroundPopup_Closed);

			FBackgroundPopup = new ColorPopup();
			FBackgroundPopup.Closed += new EventHandler(backgroundPopup_Closed);
		}

		//---------------------------------------------------------------------------------------------
		private void Window_Closing(object sender, CancelEventArgs e)
		{
			FForegroundPopup.Closed -= new EventHandler(foregroundPopup_Closed);
			FBackgroundPopup.Closed -= new EventHandler(backgroundPopup_Closed);

			// 現在のフォームの位置とサイズを初期化ファイルに記録する
			var ini = new emanual.Wpf.Utility.XmlIni(INI_FILE);
			ini.WriteFormPositionAndSize(this, GENERAL_SEC, FORM_POS_KEY);
		}

		//---------------------------------------------------------------------------------------------
		private void mnuFileOpen_Click(object sender, RoutedEventArgs e)
		{
			var dlg = new Microsoft.Win32.OpenFileDialog();
			dlg.Filter = "XSHD ファイル|*.xshd|XML ファイル|*.xml|すべてのファイル|*.*";
			dlg.FileName = "*.xshd";

			string path = System.Reflection.Assembly.GetEntryAssembly().Location;
			dlg.InitialDirectory = System.IO.Path.GetDirectoryName(path);

			if (dlg.ShowDialog() == true)
			{
				FXshdFileName = dlg.FileName;

				this.SetupDataList(FXshdFileName);
				statusItem1.Content= FXshdFileName;
			}
		}

		//---------------------------------------------------------------------------------------------
		private void mnuFileSave_Click(object sender, RoutedEventArgs e)
		{
			if (FDocument == null)
				return;

			this.SaveDocumentToFile(FXshdFileName);
		}

		//---------------------------------------------------------------------------------------------
		private void mnuFileSaveAs_Click(object sender, RoutedEventArgs e)
		{
			if (String.IsNullOrEmpty(FXshdFileName))
			{
				MessageBox.Show("先に XSHD ファイルを読み込んでください。");
				return;
			}

			var dlg = new Microsoft.Win32.SaveFileDialog();
			dlg.Filter = "XSHD ファイル|*.xshd|XML ファイル|*.xml|すべてのファイル|*.*";
			dlg.FileName = System.IO.Path.GetFileName(FXshdFileName);
			dlg.AddExtension = true;
			dlg.DefaultExt = "xshd";

			if (dlg.ShowDialog() == true)
			{
				FXshdFileName = dlg.FileName;
				statusItem1.Content = FXshdFileName;

				this.SaveDocumentToFile(FXshdFileName);
			}
		}

		//---------------------------------------------------------------------------------------------
		private void mnuFileExit_Click(object sender, RoutedEventArgs e)
		{
			this.Close();
		}

		//---------------------------------------------------------------------------------------------
		private void mnuShowBackground_Click(object sender, RoutedEventArgs e)
		{
			if (String.IsNullOrEmpty(FXshdFileName))
			{
				MessageBox.Show("先に XSHD ファイルを読み込んでください。");
				return;
			}

			var dlg = new emanual.Wpf.Controls.ColorPicker();
			dlg.OwnerWindow = this;

			if (dlg.ShowDialog() == true)
			{
				FEditorBackground = new SolidColorBrush(dlg.SelectedColor);

				for (int i = 0; i < FDataCollection.Count; ++i)
				{
					var data = FDataCollection[i].Clone();
					data.EditorBackground = FEditorBackground;
					FDataCollection[i] = data;
				}

				var ini = new emanual.Wpf.Utility.XmlIni(INI_FILE);
				ini.WriteValue(GENERAL_SEC, BACKGROUND_KEY, FEditorBackground);
			}
		}

		//---------------------------------------------------------------------------------------------
		private void mnuShowForeground_Click(object sender, RoutedEventArgs e)
		{
			if (String.IsNullOrEmpty(FXshdFileName))
			{
				MessageBox.Show("先に XSHD ファイルを読み込んでください。");
				return;
			}

			var dlg = new emanual.Wpf.Controls.ColorPicker();
			dlg.OwnerWindow = this;

			if (dlg.ShowDialog() == true)
			{
				FEditorForeground = new SolidColorBrush(dlg.SelectedColor);

				for (int i = 0; i < FDataCollection.Count; ++i)
				{
					var data = FDataCollection[i].Clone();
					data.EditorForeground = FEditorForeground;
					FDataCollection[i] = data;
				}

				var ini = new emanual.Wpf.Utility.XmlIni(INI_FILE);
				ini.WriteValue(GENERAL_SEC, FOREGROUND_KEY, FEditorForeground);
			}
		}

		//---------------------------------------------------------------------------------------------
		private void btnForeground_Click(object sender, RoutedEventArgs e)
		{
			Button button = sender as Button;

			if (button != null)
			{
				FButtonTag = (int)button.Tag;

				var binding = new System.Windows.Data.Binding();
				binding.Source = button;
				FForegroundPopup.SetBinding(System.Windows.Controls.Primitives.Popup.PlacementTargetProperty, binding);
				FForegroundPopup.IsOpen = true; // 以後の処理は foregroundPopup_Closed で実行する
			}
		}

		//---------------------------------------------------------------------------------------------
		private void foregroundPopup_Closed(object sender, EventArgs e)
		{
			if (FForegroundPopup.SelectedColor.A == 0) // ポップアップをキャンセルすると、#00000000 が戻る
				return;

			var data = FDataCollection[FButtonTag].Clone();
			data.Foreground = new SolidColorBrush(FForegroundPopup.SelectedColor);

			FDataCollection[FButtonTag] = data;

			this.SaveDocumentToFile(FXshdFileName);
			this.SetupDataList(FXshdFileName);
		}

		//---------------------------------------------------------------------------------------------
		private void btnBackground_Click(object sender, RoutedEventArgs e)
		{
			Button button = sender as Button;

			if (button != null)
			{
				FButtonTag = (int)button.Tag;

				var binding = new System.Windows.Data.Binding();
				binding.Source = button;
				FBackgroundPopup.SetBinding(System.Windows.Controls.Primitives.Popup.PlacementTargetProperty, binding);
				FBackgroundPopup.IsOpen = true; // 以後の処理は backgroundPopup_Closed で実行する
			}
		}

		//---------------------------------------------------------------------------------------------
		private void backgroundPopup_Closed(object sender, EventArgs e)
		{
			if (FBackgroundPopup.SelectedColor.A == 0) // ポップアップをキャンセルすると、#00000000 が戻る
				return;

			var data = FDataCollection[FButtonTag].Clone();
			data.Background = new SolidColorBrush(FBackgroundPopup.SelectedColor);

			FDataCollection[FButtonTag] = data;
		}

		//---------------------------------------------------------------------------------------------
		// 指定のファイルに XSHD ドキュメントの内容を保存する
		private void SaveDocumentToFile(string fileName)
		{
			// Color 要素のリスト
			XmlNodeList list = FDocument.GetElementsByTagName("Color");

			if ((list == null) || (list.Count < 1))
				return;

			for (int i = 0; i < list.Count; ++i)
			{
				var attrs = list[i].Attributes;

				//----------------------------------------------
				var foregroundAttr = attrs["foreground"];

				if (foregroundAttr != null)
					foregroundAttr.InnerXml = FDataCollection[i].ForegroundName;
				else
				{
					var attr = FDocument.CreateAttribute("foreground");
					attr.InnerXml = FDataCollection[i].ForegroundName;
					attrs.Append(attr);
				}

				//----------------------------------------------
				var backgroundAttr = attrs["background"];

				if (backgroundAttr != null)
					backgroundAttr.InnerXml = FDataCollection[i].BackgroundName;
				else if (String.Compare(FDataCollection[i].BackgroundName, "transparent", true) != 0)
				{
					var attr = FDocument.CreateAttribute("background");
					attr.InnerXml = FDataCollection[i].BackgroundName;
					attrs.Append(attr);
				}

				//----------------------------------------------
				var fontStyleAttr = attrs["fontStyle"];

				if (fontStyleAttr != null)
				{
					if (FDataCollection[i].FontStyle != FontStyles.Normal)
						fontStyleAttr.InnerXml = FDataCollection[i].FontStyle.ToString().ToLower();
					else
						attrs.Remove(fontStyleAttr);
				}
				else
				{
					if (FDataCollection[i].FontStyle != FontStyles.Normal)
					{
						var attr = FDocument.CreateAttribute("fontStyle");
						attr.InnerXml = FDataCollection[i].FontStyle.ToString().ToLower();
						attrs.Append(attr);
					}
					else // Normal のとき、属性を削除する
						attrs.Remove(fontStyleAttr);
				}

				//----------------------------------------------
				var fontWeightAttr = attrs["fontWeight"];

				if (fontWeightAttr != null)
				{
					if (FDataCollection[i].FontWeight != FontWeights.Normal)
						fontWeightAttr.InnerXml = FDataCollection[i].FontWeight.ToString().ToLower();
					else
						attrs.Remove(fontWeightAttr);
				}
				else
				{
					if (FDataCollection[i].FontWeight != FontWeights.Normal)
					{
						var attr = FDocument.CreateAttribute("fontWeight");
						attr.InnerXml = FDataCollection[i].FontWeight.ToString().ToLower();
						attrs.Append(attr);
					}
					else // Normal のとき、属性を削除する
						attrs.Remove(fontWeightAttr);
				}
			}
			try
			{
				FDocument.Save(fileName);
			}
			catch (Exception ex)
			{
				MessageBox.Show(ex.Message);
			}
		}

		//---------------------------------------------------------------------------------------------
		// コード例用 TextEditor コントロールのためのハイライト表示オブジェクトを取得する
		private IHighlightingDefinition GetHighlightDefinition(string fileName)
		{
			var reader = new System.Xml.XmlTextReader(fileName);
			var definition = ICSharpCode.AvalonEdit.Highlighting.Xshd.HighlightingLoader.Load(reader, HighlightingManager.Instance);
			reader.Close();

			return definition;
		}

		//---------------------------------------------------------------------------------------------
		// 指定の XSHD ファイルを読み込んで、FDataCollection を構築する
		private void SetupDataList(string fileName)
		{
			if (String.IsNullOrEmpty(fileName))
				throw new ArgumentNullException("XSHD ファイル名");

			if (System.IO.File.Exists(fileName) == false)
			{
				MessageBox.Show("指定の XSHD ファイルは存在しません。");
				return;
			}

			FDataCollection = new ObservableCollection<XshdData>();

			try
			{
				FDocument = new System.Xml.XmlDocument();
				FDocument.Load(FXshdFileName);

				// Color 要素のノードリストを取得する
				XmlNodeList nodeList = FDocument.GetElementsByTagName("Color");
				int index = 0;
				IHighlightingDefinition definition = this.GetHighlightDefinition(fileName);

				foreach (XmlElement element in nodeList)
				{
					if (element.HasAttributes)
					{
						var data = new XshdData();
						data.EditorBackground = FEditorBackground;
						data.EditorForeground = FEditorForeground;
						data.Background = Brushes.Transparent;
						data.Foreground = FEditorForeground;
						data.ExampleText = String.Empty;
						data.HighlightingDefinition = definition;

						foreach (XmlAttribute attr in element.Attributes)
						{
							if (String.Compare(attr.LocalName, "name") == 0)
								data.Name = attr.Value;

							if (String.Compare(attr.LocalName, "foreground") == 0) // foreground 属性
							{
								Color color = (Color)ColorConverter.ConvertFromString(attr.Value);
								data.Foreground = new SolidColorBrush(color);
							}

							if (String.Compare(attr.LocalName, "background") == 0) // background 属性
							{
								Color color = (Color)ColorConverter.ConvertFromString(attr.Value);
								data.Background = new SolidColorBrush(color);
							}

							if (String.Compare(attr.LocalName, "fontStyle") == 0) // FontStyle 属性
							{
								string style = attr.Value as string;
								style = style.ToLower();

								if (style == "italic")
									data.FontStyle = FontStyles.Italic;
								else if (style == "oblique")
									data.FontStyle = FontStyles.Oblique;
								else
									data.FontStyle = FontStyles.Normal;
							}

							if (String.Compare(attr.LocalName, "fontWeight") == 0) // FontWeight 属性
							{
								string weight = attr.Value as string;
								weight = weight.ToLower();

								if (weight == "bold")
									data.FontWeight = FontWeights.Bold;
								else
									data.FontWeight = FontWeights.Normal;
							}

							if (String.Compare(attr.LocalName, "exampleText") == 0) // exampleTex 属性
							{
								data.ExampleText = attr.Value;
							}
						} // end of foreach attr

						data.Tag = index; // 選択したボタンを識別するため
						FDataCollection.Add(data);
						++index;
					} // end of if
				} // end of foreach element
			}
			catch (Exception ex)
			{
				MessageBox.Show(ex.Message);
			}

			dataGrid.ItemsSource = FDataCollection;
			dataGrid.SelectedIndex = 0;
		}

		//---------------------------------------------------------------------------------------------
	} // end of MainWindow class

	//***********************************************************************************************
	internal class FontStyleData : ObservableCollection<FontStyle>
	{
		public FontStyleData()
		{
			this.Add(FontStyles.Normal);
			this.Add(FontStyles.Italic);
			this.Add(FontStyles.Oblique);
		}
	}

	//***********************************************************************************************
	internal class FontWeightData : ObservableCollection<FontWeight>
	{
		public FontWeightData()
		{
			this.Add(FontWeights.Normal);
			this.Add(FontWeights.Bold);
		}
	}

	//***********************************************************************************************
	internal class XshdData : DependencyObject
	{
		private string FName;
		private SolidColorBrush FEditorBackground; // TextEditor コントロールの背景色
		private SolidColorBrush FEditorForeground; // 同、文字色
		private SolidColorBrush FBackground;       // テキストの背景色
		private SolidColorBrush FForeground;       // 同、文字色
		private string FForegroundName;
		private string FBackgroundName;
		private FontStyle FStyle = FontStyles.Normal;
		private FontWeight FWeight = FontWeights.Normal;
		private int FTag;
		private string FExampleText;
		private IHighlightingDefinition FHighlightingDefinition;

		public string Name { get { return FName; } set { FName = value; } }
		public SolidColorBrush EditorBackground { get { return FEditorBackground; } set { FEditorBackground = value; } }
		public SolidColorBrush EditorForeground { get { return FEditorForeground; } set { FEditorForeground = value; } }
		public SolidColorBrush Background { get { return FBackground; } set { FBackground = value; this.SetBackgroundName(value); } }
		public SolidColorBrush Foreground { get { return FForeground; } set { FForeground = value; this.SetForegroundName(value); } }
		public string ForegroundName { get { return FForegroundName; } set { FForegroundName = value; } }
		public string BackgroundName { get { return FBackgroundName; } set { FBackgroundName = value; } }
		public FontStyle FontStyle { get { return FStyle; } set { FStyle = value; } }
		public FontWeight FontWeight { get { return FWeight; } set { FWeight = value; } }
		public int Tag { get { return FTag; } set { FTag = value; } }
		public string ExampleText { get { return FExampleText; } set { FExampleText = value; this.SetTextDocumentProperty(value);} }
		public IHighlightingDefinition HighlightingDefinition { get { return FHighlightingDefinition; } set { FHighlightingDefinition = value; } }

    public static readonly DependencyProperty TextDocumentProperty;

    // プロパティ
		public TextDocument TextDocument
    {
			get { return (TextDocument)this.GetValue(TextDocumentProperty); }
			set { this.SetValue(TextDocumentProperty, value); }
    }

    //-----------------------------------------------------------------------------------
    static XshdData()
    {
			TextDocumentProperty = DependencyProperty.Register("TextDocument", typeof(TextDocument), typeof(XshdData),
          new FrameworkPropertyMetadata(null));
    }

		//---------------------------------------------------------------------------------------------
		public XshdData()
		{
		}

		//---------------------------------------------------------------------------------------------
		public XshdData Clone()
		{
			var data = new XshdData();
			data.Name = FName;
			data.EditorBackground = FEditorBackground;
			data.EditorForeground = FEditorForeground;
			data.Background = FBackground;
			data.Foreground = FForeground;
			data.ForegroundName = FForegroundName;
			data.BackgroundName = FBackgroundName;
			data.ExampleText = this.ExampleText;
			data.TextDocument = this.TextDocument;
			data.FontStyle = FStyle;
			data.FontWeight = FWeight;
			data.Tag = FTag;
			data.HighlightingDefinition = FHighlightingDefinition;

			return data;
		}

		//---------------------------------------------------------------------------------------------
		public override string ToString()
		{
			return FForegroundName;
		}

		//---------------------------------------------------------------------------------------------
		// ExampleText プロパティに基づいて TextDocument プロパティを設定する
		private void SetTextDocumentProperty(string text)
		{
			this.TextDocument = new TextDocument(text);
		}

		//---------------------------------------------------------------------------------------------
		// Foreground プロパティの設定に基づいて、ForegroundName プロパティを設定する
		private void SetForegroundName(SolidColorBrush brush)
		{
			if (brush != null)
				FForegroundName = this.GetKnownColorName(brush.Color);
			else
				FForegroundName = "ControlTextColor"; // 事実上、ここにはこない
		}

		//---------------------------------------------------------------------------------------------
		// Background プロパティの設定に基づいて、BackgroundName プロパティを設定する
		private void SetBackgroundName(SolidColorBrush brush)
		{
			if (brush != null)
				FBackgroundName = this.GetKnownColorName(brush.Color);
			else
				FBackgroundName = "Window"; // 事実上、ここにはこない
		}

		//---------------------------------------------------------------------------------------------
		// Color 型から Colors クラスのプロパティ名を取得する
		// 該当するものがない場合は #AARRGGBB タイプの文字列を返す
		private string GetKnownColorName(Color color)
		{
			Type type = typeof(System.Windows.Media.Colors);
			BindingFlags flags = BindingFlags.Static | BindingFlags.Public;

			PropertyInfo[] infos = type.GetProperties(flags);

			int index = -1;
			string s = color.ToString();

			for (int i = 0; i < infos.Length; i++)
			{
				Color c = (Color)infos[i].GetValue(null, null);

				if (c == color)
				{
					index = i;
					break;
				}
			}

			if (index != -1)
				s = infos[index].Name;

			return s;
		}
	} // end of XshdData class
} // end of namespace
